import math
import time
import numpy as np
import numbers
from scipy.optimize import linprog
import matplotlib.pyplot as plt

import SInterval as SInt
import FNumber as FN

#import pacal as PAC
#PAC.params.general.parallel = False


############################################################################################
# def Segments2Distr(pdf_seg):
#     '''
#     Function converts PiecewiseDistribution object to lists of arguments and pdf values.
#     Functions works for discrete distributions.
#     pdf_seg: PiecewiseDistribution object
#     returns: a - list of arguments
#              f - list of pdf values
#
#     Example:
#         M = PAC.DiscreteDistr(xi=[1.0, 2.0, 3.0, 4.0], pi=[0.3, 0.4, 0.2, 0.1])
#         p = M.get_piecewise_pdf()
#         a1,f1 = Segments2Distr(p)
#     '''
#     a = []
#     f = []
#     for seg in pdf_seg.segments:
#         if  seg.isDirac():
#             a.append(seg.a)
#             f.append(seg.f)
#     return a,f



############################################################################################
def calc_discrete_pdf(x1, px1, x2, px2, op='+'):
    '''
    Calculates result of calculations on discrete pdf's.
    x1, px1, x2, px2: arguments and pdf's of probabilistic discrete number 1 and 2
    op: '+', '-', '*', '/', 'a' (for abs), 'p' (for power)
    return y, yp: arguments and pdf of result
    Example:
        x1 = [1.0, 2.0, 3.0, 4.0, 5.0]
        px1 = [0.1, 0.3, 0.3, 0.2, 0.1]
        x2 = [2.0, 3.0, 4.0, 5.0]
        px2 = [0.2, 0.3, 0.3, 0.2]
        Sy,Syp = calc_discrete_pdf(x1, px1, x2, px2, op='+')
    '''

    n1 = len(x1)
    n2 = len(x2)
    v = np.zeros(n1*n2)
    vp = np.zeros(n1*n2)

    k = 0
    for i in range(n1):
        for j in range(n2):
            if  op == '+':
                v[k] = x1[i] + x2[j]
            elif op == '-':
                v[k] = x1[i] - x2[j]
            elif op == '*':
                v[k] = x1[i] * x2[j]
            elif op == '/':
                v[k] = x1[i] / x2[j]
            elif op == 'a':
                v[k] = abs(x1[i])
            elif op == 'p':
                v[k] = x1[i] ** x2[j]

            vp[k] = px1[i] * px2[j]
            k += 1

    ind = np.argsort(v)
    v = v[ind]
    vp = vp[ind]

    y = np.zeros(n1 * n2)
    yp = np.zeros(n1 * n2)
    licz = 0
    y[0] = v[0]
    yp[0] = vp[0]
    v_last = v[0]
    for k in range(1,n1*n2):
        if  v[k] == v_last:
            yp[licz] += vp[k]
        else:
            licz += 1
            y[licz] = v[k]
            yp[licz] = vp[k]
            v_last = v[k]

    y = np.delete(y, range(licz+1,n1*n2), axis=0)
    yp = np.delete(yp, range(licz + 1, n1 * n2), axis=0)
    return y, yp



############################################################################################
class Parameters:
    interp_points = 20  # number of interpolation points
    discrete_points = 5 # nummer of point for discrete membership function of probability
    # sim_interp_points = 100 # number of interpolation points in similarity measure calculations



############################################################################################
def operation(in1, in2, op = '+'):
    if  isinstance(in2, numbers.Number):
        b_f = FN.Singleton(in2, '{0}'.format(in2))
        b_p = FN.Singleton(1.0, '1')
        b = ZNumber(b_f, b_p)
    elif isinstance(in2, SInt.SInterval):
        b_f = FN.Rectangle([in2.min_val, in2.max_val], '[{0.min_val},{0.max_val}]'.format(in2))
        b_p = FN.Singleton(1.0, '1')
        b = ZNumber(b_f, b_p)
    elif isinstance(in2, FN.FNumber):
        b_f = in2
        b_p = FN.Singleton(1.0, '1')
        b = ZNumber(b_f, b_p)
    elif isinstance(in2, ZNumber):
        b  = in2
    else:
        raise NotImplemented()

    a = in1
    if  op == '+':
        c_fn = a.FNum + b.FNum
    elif op == '-':
        c_fn = a.FNum - b.FNum
    elif op == '*':
        c_fn = a.FNum * b.FNum
    elif op == '/':
        c_fn = a.FNum / b.FNum
    elif op == 'a':
        c_fn = abs(a.FNum)
    elif op == 'p':
        c_fn = a.FNum ** b.FNum.singleton_position

    a_s = a.FNum.support()
    a_ds = np.linspace(a_s.min_val, a_s.max_val, Parameters.interp_points + 1)
    a_ds_mu = np.zeros_like(a_ds)
    for i in range(Parameters.interp_points+1):
        a_ds_mu[i] = a.FNum.muValue(a_ds[i])

    ap_ds = np.zeros(2*(Parameters.discrete_points + 1))
    ap_ds_mu = np.zeros_like(ap_ds)
    i = 0
    for alfa in np.linspace(0.0, 1.0, Parameters.discrete_points + 1):
        r = a.PVal.alphaValue(alfa)
        ap_ds[i] = r.min_val
        ap_ds[-1-i] = r.max_val
        ap_ds_mu[i] = alfa
        ap_ds_mu[-1 - i] = alfa
        i += 1

    A_mat = [a_ds_mu, np.ones_like(a_ds_mu)]
    c_mat = np.zeros_like(a_ds_mu)
    ap_distr = []
    for i in range(2*(Parameters.discrete_points + 1)):
        b_mat = [ap_ds[i], 1]
        res = linprog(c_mat, A_eq=A_mat, b_eq=b_mat, bounds=(0, 1), options={"disp": False}, method='interior-point')  # method='simplex'   method='interior-point'
        ap_distr.append(res.x)

    b_s = b.FNum.support()
    b_ds = np.linspace(b_s.min_val, b_s.max_val, Parameters.interp_points + 1)
    b_ds_mu = np.zeros_like(b_ds)
    for i in range(Parameters.interp_points + 1):
        b_ds_mu[i] = b.FNum.muValue(b_ds[i])

    bp_ds = np.zeros(2*(Parameters.discrete_points + 1))
    bp_ds_mu = np.zeros_like(bp_ds)
    i = 0
    for alfa in np.linspace(0.0, 1.0, Parameters.discrete_points + 1):
        r = b.PVal.alphaValue(alfa)
        bp_ds[i] = r.min_val
        bp_ds[-1-i] = r.max_val
        bp_ds_mu[i] = alfa
        bp_ds_mu[-1 - i] = alfa
        i += 1

    A_mat = [b_ds_mu, np.ones_like(b_ds_mu)]
    c_mat = np.zeros_like(b_ds_mu)
    bp_distr = []
    for i in range(2*(Parameters.discrete_points + 1)):
        b_mat = [bp_ds[i], 1]
        res = linprog(c_mat, A_eq=A_mat, b_eq=b_mat, bounds=(0, 1), options={"disp": False}, method='interior-point')  # method='simplex'   method='interior-point'
        bp_distr.append(res.x)

    prob = []
    prob_mu = []
    for i in range(len(ap_ds)):
        for j in range(len(bp_ds)):
            a1, f1 = calc_discrete_pdf(a_ds, ap_distr[i], b_ds, bp_distr[j], op)
            suma = 0
            for k in range(len(a1)):
                suma += c_fn.muValue(a1[k]) * f1[k]
            prob.append(suma)
            prob_mu.append(min(ap_ds_mu[i],bp_ds_mu[j]))

    #plt.figure(num=10); plt.subplot(1,2,2); plt.plot(prob,prob_mu,'o',MarkerSize=5)

    c_fn_prob = FN.FNumber()
    c_fn_prob.fntype = 0
    c_fn_prob.name = ' '
    c_fn_prob.mu = np.linspace(0.0, 1.0, Parameters.discrete_points + 1)
    c_fn_prob.left = np.zeros(Parameters.discrete_points + 1)
    c_fn_prob.right = np.zeros(Parameters.discrete_points + 1)
    k = 0
    for alfa in c_fn_prob.mu:
        pierwszy = True
        for i in range(len(prob)):
            if prob_mu[i] == alfa:
                if  pierwszy:
                    min1 = prob[i]
                    max1 = prob[i]
                    pierwszy = False
                else:
                    if  prob[i] > max1:
                        max1 = prob[i]
                    if  prob[i] < min1:
                        min1 = prob[i]
        c_fn_prob.left[k] = min1
        c_fn_prob.right[k] = max1
        k += 1
        if  alfa == 1.0:
            c_fn_prob.singleton_position = 0.5*(min1+max1)

    return ZNumber(c_fn, c_fn_prob)



############################################################################################
class ZNumber:
    '''
    Z Number Class
    '''

    def __init__(self, fuz_number=FN.Rectangle([0, 1], 'Interval [0,1]'), prob_number=FN.Rectangle([0.5, 1], 'Big')):

        if  isinstance(fuz_number, numbers.Number):
            fn = FN.Singleton(fuz_number, '{0}'.format(fuz_number))
        elif isinstance(fuz_number, SInt.SInterval):
            fn = FN.Rectangle([fuz_number.min_val, fuz_number.max_val], '[{0.min_val},{0.max_val}]'.format(fuz_number))
        else:
            fn = fuz_number

        if  isinstance(prob_number, numbers.Number):
            pv = FN.Singleton(prob_number, '{0}'.format(prob_number))
        elif isinstance(fuz_number, SInt.SInterval):
            pv = FN.Rectangle([prob_number.min_val, prob_number.max_val], '[{0.min_val},{0.max_val}]'.format(prob_number))
        else:
            pv = prob_number

        self.FNum = fn
        self.PVal = pv


    def __repr__(self):
        return self.__str__()


    def __str__(self):
        return self.FNum.__str__() + ' with ' + self.PVal.__str__() + ' probability'


    def plot(self):
        plt.subplot(1, 2, 1)
        self.FNum.plot()
        # plt.title('Value')
        plt.title('Wartość')
        plt.grid(True)
        plt.subplot(1, 2, 2)
        self.PVal.plot()
        # plt.title('Probability')
        plt.title('Prawdopodobieństwo')
        plt.grid(True)

        return plt.gca()


    def __neg__(self):
        return (ZNumber(-self.FNum, self.PVal))


    def __add__(self, other):
        if  isinstance(other, numbers.Number):
            return ZNumber(self.FNum + other, self.PVal)
        else:
            return operation(self, other, op='+')


    def __radd__(self, other):
        return self.__add__(other)


    def __sub__(self, other):
        if  isinstance(other, numbers.Number):
            return ZNumber(self.FNum - other, self.PVal)
        else:
            return operation(self, other, op='-')


    def __rsub__(self, other):
        return other + (-self)


    def __mul__(self, other):
        if  isinstance(other, numbers.Number):
            return ZNumber(self.FNum * other, self.PVal)
        else:
            return operation(self, other, op='*')


    def __rmul__(self, other):
        return self.__mul__(other)


    def __truediv__(self, other):
        if  isinstance(other, numbers.Number):
            if  other == 0.0:
                raise ValueError('Division by 0!')
            return ZNumber(self.FNum / other, self.PVal)
        else:
            return operation(self, other, op='/')


    def __rtruediv__(self, other):
        if  isinstance(other, numbers.Number):
            z = ZNumber(FN.Singleton(other, '{0}'.format(other)), FN.Singleton(1.0, '1'))
            return z/self
        elif isinstance(other, SInt.SInterval):
            z_f = FN.Rectangle([other.min_val, other.max_val], '[{0.min_val},{0.max_val}]'.format(other))
            z_p = FN.Singleton(1.0, '1')
            z = ZNumber(z_f, z_p)
            return z / self
        elif isinstance(other, FN.FNumber):
            z = ZNumber(other, FN.Singleton(1.0, '1'))
            return z / self
        else:
            raise NotImplemented()


    def __abs__(self):
        s = self.FNum.support()
        if s.min_val >= 0:
            return self
        elif s.max_val <= 0:
            return -self
        else:
            return operation(self, 1.0, op='a')


    def __pow__(self, other):
        if not isinstance(other, numbers.Number):
            raise ValueError("Exponent must be a number!")

        s = self.FNum.support()
        if other < 0 and s.isSubset(0.0):
            raise ValueError("Fuzzy number cannot be raised to a negative power because its support contains 0.0")

        return operation(self, other, op='p')




##############################################################################################
# Distance between 2 Z numbers
# fn_dist_type = 1: distance between singleton positions
# fn_dist_type = 2: distance between centers of gravity
# fn_dist_type = 3: d = (\int (mu_a(x) - mu_b(x))**p dx)**(1.0/p)
# fn_dist_type = 4: integral formula with p exponent
# fn_dist_type = 5: integral formula with max
##############################################################################################
def distance(z1, z2, type=2, lambda_A=1, fn_dist_type=1, p=2):
    d_fn = FN.distance(z1.FNum, z2.FNum, fn_dist_type, p)
    d_pd = FN.distance(z1.PVal, z2.PVal, fn_dist_type, p)

    if  type == 1:
        return lambda_A*d_fn + d_pd
    elif type == 2:
        return np.sqrt(lambda_A*(d_fn**2) + d_pd**2)





